'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Traverser = function () {
  function Traverser(tree) {
    (0, _classCallCheck3.default)(this, Traverser);

    this._tree = tree;
  }

  /**
    * Searches a tree in DFS fashion. Requires a search criteria to be provided.
    *
    * @method searchDFS
    * @memberof Traverser
    * @instance
    * @param {function} criteria - MUST BE a callback function that specifies the search criteria.
    * Criteria callback here receives {@link TreeNode#_data} in parameter and MUST return boolean
    * indicating whether that data satisfies your criteria.
    * @return {object} - first {@link TreeNode} in tree that matches the given criteria.
    * @example
    * // Search DFS
    * var node = tree.searchDFS(function(data){
    *  return data.key === '#greenapple';
    * });
    */


  (0, _createClass3.default)(Traverser, [{
    key: 'searchDFS',
    value: function searchDFS(criteria) {

      // Hold the node when found
      var foundNode = null;

      // Find node recursively
      (function recur(node) {
        if (node.matchCriteria(criteria)) {
          foundNode = node;
          return foundNode;
        } else {
          node._childNodes.some(recur);
        }
      })(this._tree._rootNode);

      return foundNode;
    }

    /**
     * Searches a tree in BFS fashion. Requires a search criteria to be provided.
     *
     * @method searchBFS
     * @memberof Traverser
     * @instance
     * @param {function} criteria - MUST BE a callback function that specifies the search criteria.
     * Criteria callback here receives {@link TreeNode#_data} in parameter and MUST return boolean
     * indicating whether that data satisfies your criteria.
     * @return {object} - first {@link TreeNode} in tree that matches the given criteria.
     * @example
     * // Search BFS
     * var node = tree.searchBFS(function(data){
     *  return data.key === '#greenapple';
     * });
     */

  }, {
    key: 'searchBFS',
    value: function searchBFS(criteria) {

      // Hold the node when found
      var foundNode = null;

      // Find nodes recursively
      (function expand(queue) {
        while (queue.length) {
          var current = queue.splice(0, 1)[0];
          if (current.matchCriteria(criteria)) {
            foundNode = current;
            return;
          }
          current._childNodes.forEach(function (_child) {
            queue.push(_child);
          });
        }
      })([this._tree._rootNode]);

      return foundNode;
    }

    /**
     * Traverses an entire tree in DFS fashion.
     *
     * @method traverseDFS
     * @memberof Traverser
     * @instance
     * @param {function} callback - Gets triggered when @{link TreeNode} is explored. Explored node is passed as parameter to callback.
     * @example
     * // Traverse DFS
     * tree.traverseDFS(function(node){
     *  console.log(node.data);
     * });
     */

  }, {
    key: 'traverseDFS',
    value: function traverseDFS(callback) {
      (function recur(node) {
        callback(node);
        node._childNodes.forEach(recur);
      })(this._tree._rootNode);
    }

    /**
     * Traverses an entire tree in BFS fashion.
     *
     * @method traverseBFS
     * @memberof Traverser
     * @instance
     * @param {function} callback - Gets triggered when node is explored. Explored node is passed as parameter to callback.
     * @example
     * // Traverse BFS
     * tree.traverseBFS(function(node){
     *  console.log(node.data);
     * });
     */

  }, {
    key: 'traverseBFS',
    value: function traverseBFS(callback) {
      (function expand(queue) {
        while (queue.length) {
          var current = queue.splice(0, 1)[0];
          callback(current);
          current._childNodes.forEach(function (_child) {
            queue.push(_child);
          });
        }
      })([this._tree._rootNode]);
    }
  }]);
  return Traverser;
}();

var TreeNode = function () {
  function TreeNode(data) {
    (0, _classCallCheck3.default)(this, TreeNode);

    /**
       * Represents the parent node
       *
       * @property _parentNode
       * @type {object}
       * @default "null"
       */
    this._parentNode = null;

    /**
     * Represents the child nodes
     *
     * @property _childNodes
     * @type {array}
     * @default "[]"
     */
    this._childNodes = [];

    /**
     * Represents the data node has
     *
     * @property _data
     * @type {object}
     * @default "null"
     */
    this._data = data;

    /**
     * Depth of the node represents level in hierarchy
     *
     * @property _depth
     * @type {number}
     * @default -1
     */
    this._depth = -1;
  }

  (0, _createClass3.default)(TreeNode, [{
    key: 'parentNode',
    value: function parentNode() {
      return this._parentNode;
    }
  }, {
    key: 'childNodes',
    value: function childNodes() {
      return this._childNodes;
    }
  }, {
    key: 'data',
    value: function data(_data2) {
      if (arguments.length > 0) {
        this._data = _data2;
      } else {
        return this._data;
      }
    }
  }, {
    key: 'depth',
    value: function depth() {
      return this._depth;
    }
  }, {
    key: 'matchCriteria',
    value: function matchCriteria(criteria) {
      return criteria(this._data);
    }

    /**
      * get sibling nodes.
      *
      * @method siblings
      * @memberof TreeNode
      * @instance
      * @return {array} - array of instances of {@link TreeNode}
      */

  }, {
    key: 'siblings',
    value: function siblings() {
      var _this = this;

      return !this._parentNode ? [] : this._parentNode._childNodes.filter(function (_child) {
        return _child !== _this;
      });
    }

    /**
     * Gets an array of all ancestor nodes
     *
     * @method getAncestry
     * @memberof TreeNode
     * @instance
     * @return {Array} - array of ancestor nodes
     */

  }, {
    key: 'getAncestry',
    value: function getAncestry() {
      // Initialize empty array and node
      var ancestors = [],
          node = this;

      // Loop over ancestors and add them in array
      while (node.parentNode()) {
        ancestors.push(node.parentNode());
        node = node.parentNode();
      }

      return ancestors;
    }
  }, {
    key: 'export',
    value: function _export() {
      var criteria = arguments.length <= 0 || arguments[0] === undefined ? function (data) {
        return data;
      } : arguments[0];

      // Check if criteria is specified
      if (!criteria || typeof criteria !== 'function') throw new Error('Export criteria not specified');

      // Export every node recursively
      var exportRecur = function exportRecur(node) {
        var exported = node.matchCriteria(criteria);

        if (!exported || (typeof exported === 'undefined' ? 'undefined' : (0, _typeof3.default)(exported)) !== 'object') {
          throw new Error('Export criteria should always return an object and it cannot be null.');
        } else {
          exported.children = [];
          node._childNodes.forEach(function (_child) {
            exported.children.push(exportRecur(_child));
          });

          return exported;
        }
      };

      return exportRecur(this);
    }
  }]);
  return TreeNode;
}();

var Tree = function () {
  function Tree() {
    (0, _classCallCheck3.default)(this, Tree);

    this._rootNode = null;
    this._currentNode = null;
    this._traverser = new Traverser(this);
  }

  (0, _createClass3.default)(Tree, [{
    key: 'traverser',
    value: function traverser() {
      return this._traverser;
    }
  }, {
    key: 'searchBFS',
    value: function searchBFS(criteria) {
      return this.traverser().searchBFS(criteria);
    }
  }, {
    key: 'searchDFS',
    value: function searchDFS(criteria) {
      return this.traverser().searchDFS(criteria);
    }
  }, {
    key: 'traverseBFS',
    value: function traverseBFS(callback) {
      return this.traverser().traverseBFS(callback);
    }
  }, {
    key: 'traverseDFS',
    value: function traverseDFS(callback) {
      return this.traverser().traverseDFS(callback);
    }
  }, {
    key: 'isEmpty',
    value: function isEmpty() {
      return this._rootNode === null && this._currentNode === null;
    }
  }, {
    key: 'insert',
    value: function insert(data) {
      var node = new TreeNode(data);
      if (this.isEmpty()) {
        node._depth = 1;
        this._rootNode = this._currentNode = node;
      } else {
        node._parentNode = this._currentNode;
        this._currentNode._childNodes.push(node);
        this._currentNode = node;
        node.depth = node._parentNode._depth + 1;
      }
      return node;
    }
  }, {
    key: 'insertToNode',
    value: function insertToNode(node, data) {
      var newNode = new TreeNode(data);
      newNode._parentNode = node;
      newNode._depth = newNode._parentNode._depth + 1;
      node._childNodes.push(newNode);
      this._currentNode = newNode;
      return newNode;
    }
  }, {
    key: 'trimBranchFrom',
    value: function trimBranchFrom(node) {
      // trim brach recursively
      (function recur(node) {
        node._childNodes.forEach(recur);
        node._childNodes = [];
        node._data = null;
      })(node);

      // Update Current Node
      if (node._parentNode) {
        node._parentNode._childNodes.splice(node._parentNode._childNodes.indexOf(node), 1);
        this._currentNode = node._parentNode;
      } else {
        this._rootNode = this._currentNode = null;
      }
    }
  }, {
    key: 'export',
    value: function _export(criteria) {
      // Check if rootNode is not null
      if (!this._rootNode) {
        return null;
      }

      var treeData = this._rootNode.export(criteria);

      if (treeData['__root']) {
        return treeData.children;
      }

      return [treeData];
    }
  }, {
    key: 'import',
    value: function _import(data) {
      var _this2 = this;

      var criteria = arguments.length <= 1 || arguments[1] === undefined ? function (data) {
        return data;
      } : arguments[1];

      // Empty all tree
      if (this._rootNode) this.trimBranchFrom(this._rootNode);

      // Set Current Node to root node as null
      this._currentNode = this._rootNode = null;

      // Import recursively
      var importRecur = function importRecur(node, recurData) {

        // Format data from given criteria
        var _data = criteria(recurData);

        // Create Root Node
        if (!node) {
          node = _this2.insert(_data);
        } else {
          node = _this2.insertToNode(node, _data);
        }

        // For Every Child
        recurData.children && recurData.children.forEach(function (_child) {
          importRecur(node, _child);
        });
      };

      var treeData = void 0;

      if (Array.isArray(data)) {
        if (data.length === 1) {
          treeData = data[0];
        } else {
          treeData = { __root: !0, children: data };
        }
      } else {
        treeData = data;
      }

      importRecur(this._rootNode, treeData);

      // Set Current Node to root node
      this._currentNode = this._rootNode;

      return this;
    }
  }]);
  return Tree;
}();

exports.default = Tree;